﻿/*
 * Created by SharpDevelop.
 * User: banana
 * Date: 2013/10/23
 * Time: 17:26
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using log4net;
using PicLogic.Processors;

namespace PicLogic
{
	/// <summary>
	/// Description of Explainer.
	/// </summary>
	public class Explainer
	{
		private ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);
		public Sheet Sheet { get; set; }
		public int Step { get; set; }
		public MainForm Form { get; set; }
		
		private BaseProcessor[] processors = {
			new CheckNumberProcessor(),
			new CheckCountProcessor(),
			new CheckFixedProcessor(),
			new CheckNumberMultiProcessor(),
			new SplitProcessor(),
			new NarrowProcessor(),
			new CloseMaxNumProcessor(),
			new NotReachProcessor(),
			new OverlapProcessor(),
			new FromEdgeProcessor(),
		};
		
		private int changeCount;
		
		private bool IsStepOver {
			get {
				return Step > 0 && changeCount >= Step;
			}
		}
		
		public Explainer()
		{
		}
		
		public void Start()
		{
			try {
				DoExplain();
			}
			catch(Exception ex) {
				log.Error("ERROR!!", ex);
			}
		}
		
		private void DoExplain()
		{
			while(!IsStepOver){
				changeCount = 0;
				/* 縦のラインを見ていく */
				for(int index=0; index<Sheet.VertNumbers.Count; index++){
					if(IsStepOver) break;
					/* 上から下に */
					ContinuityNumber[] nums = Sheet.VertNumbers[index];
					Brock[] line = GetVertLine(index);
					bool changed = false;
					foreach(BaseProcessor p in processors){
						if(IsStepOver) break;
						log.DebugFormat("V{0} {1}", index, p.GetType().Name);
						if(UpdateVertLine(index, p.Do(nums, line, true))){
							changed = true;
						}
					}
					if(changed){
						changeCount++;
					}
				}
				/* 横のラインを見ていく */
				for(int index=0; index<Sheet.HorizNumbers.Count; index++){
					if(IsStepOver) break;
					/* 左から右に */
					ContinuityNumber[] nums = Sheet.HorizNumbers[index];
					Brock[] line = GetHorizLine(index);
					bool changed = false;
					foreach(BaseProcessor p in processors){
						if(IsStepOver) break;
						log.DebugFormat("H{0} {1}", index, p.GetType().Name);
						if(UpdateHorizLine(p.Do(nums, line, false), index)){
							changed = true;
						}
					}
					if(changed){
						changeCount++;
					}
				}
				if(changeCount == 0){
					log.Error("Give up!!");
					break;
				}
			}
			log.Info("End");
		}
		
		private void UpdateVertNums(int x)
		{
			int[] count = new int[Sheet.VertNumbers[x].Length];
			for(int index=0; index<Sheet.BrockVertCount; index++){
				int numIndex = Sheet.Brocks[x, index].VertNumIndex;
				if(numIndex >= 0){
					count[numIndex]++;
				}
			}
			for(int numIndex=0; numIndex<Sheet.VertNumbers[x].Length; numIndex++){
				Sheet.VertNumbers[x][numIndex].FillCount = count[numIndex];
			}
		}
		
		private void UpdateHorizNums(int y)
		{
			int[] count = new int[Sheet.HorizNumbers[y].Length];
			for(int index=0; index<Sheet.BrockHorizCount; index++){
				int numIndex = Sheet.Brocks[index, y].HorizNumIndex;
				if(numIndex >= 0){
					count[numIndex]++;
				}
			}
			for(int numIndex=0; numIndex<Sheet.HorizNumbers[y].Length; numIndex++){
				Sheet.HorizNumbers[y][numIndex].FillCount = count[numIndex];
			}
		}
		
		private bool UpdateVertLine(int x, int[] yArray)
		{
			if(yArray.Length > 0){
				UpdateVertNums(x);
			}
			foreach(int y in yArray){
				log.DebugFormat("({0},{1}) is {2}(numIndex={3},{4})",
				                x, y, Sheet.Brocks[x,y].State,
				                Sheet.Brocks[x,y].HorizNumIndex,
				                Sheet.Brocks[x,y].VertNumIndex);
				Form.UpdateBrockState(x, y);
			}
			return yArray.Length > 0;
		}
		
		private bool UpdateHorizLine(int[] xArray, int y)
		{
			if(xArray.Length > 0){
				UpdateHorizNums(y);
			}
			foreach(int x in xArray){
				log.DebugFormat("({0},{1}) is {2}(numIndex={3},{4})",
				                x, y, Sheet.Brocks[x,y].State,
				                Sheet.Brocks[x,y].HorizNumIndex,
				                Sheet.Brocks[x,y].VertNumIndex);
				Form.UpdateBrockState(x, y);
			}
			if(xArray.Length > 0){
				changeCount++;
			}
			return xArray.Length > 0;
		}
		
		private Brock[] GetVertLine(int indexX)
		{
			Brock[] result = new Brock[Sheet.BrockVertCount];
			for(int indexY=0; indexY<Sheet.BrockVertCount; indexY++){
				result[indexY] = Sheet.Brocks[indexX, indexY];
			}
			return result;
		}
		
		private Brock[] GetHorizLine(int indexY)
		{
			Brock[] result = new Brock[Sheet.BrockHorizCount];
			for(int indexX=0; indexX<Sheet.BrockHorizCount; indexX++){
				result[indexX] = Sheet.Brocks[indexX, indexY];
			}
			return result;
		}
		
		private int[] GetVertNumIndexes(Brock[] line)
		{
			int[] result = new int[Sheet.BrockVertCount];
			for(int index=0; index<result.Length; index++){
				result[index] = line[index].VertNumIndex;
			}
			return result;
		}
		
		private int[] GetHorizNumIndexes(Brock[] line)
		{
			int[] result = new int[Sheet.BrockHorizCount];
			for(int index=0; index<result.Length; index++){
				result[index] = line[index].HorizNumIndex;
			}
			return result;
		}
		
	}
}
