﻿/*
 * Created by SharpDevelop.
 * User: banana
 * Date: 2013/10/23
 * Time: 13:42
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.Threading;
using System.Windows.Forms;

namespace PicLogic
{
	/// <summary>
	/// Description of MainForm.
	/// </summary>
	public partial class MainForm : Form
	{
		private const int BROCK_PIXEL = 14;
		private const string FONT_NAME = "MS Gothic";
		private const float FONT_SIZE = 7.5F;
//		private const int BROCK_PIXEL = 5;
//		private const string FONT_NAME = "MS Gothic";
//		private const float FONT_SIZE = 5F;
		private Font _font;
		private StringFormat centeringFormat;
		private Bitmap _bitmap;
		private Sheet _sheet;
		private int _maxVertNumbersCount;
		private int _maxHorizNumbersCount;
		private Thread thread;
		private int _mouseX;
		private int _mouseY;
		
		public delegate void UpdateBrockStateDelegate(int x, int y);
		
		public MainForm()
		{
			//
			// The InitializeComponent() call is required for Windows Forms designer support.
			//
			InitializeComponent();
			
			//
			// TODO: Add constructor code after the InitializeComponent() call.
			//
			_font = new Font(FONT_NAME, FONT_SIZE);
			centeringFormat = new StringFormat() {
				Alignment = StringAlignment.Center,
				LineAlignment = StringAlignment.Center,
			};
			_mouseX = -1;
			_mouseY = -1;
		}
		
		void ToolStripButtonOpenClick(object sender, EventArgs e)
		{
			OpenFileDialog dialog = new OpenFileDialog();
			dialog.RestoreDirectory = true;
			if(dialog.ShowDialog() == DialogResult.OK){
				_sheet = ChallengeSheetReader.Read(dialog.FileName);
				InitializeSheet();
				DisplaySheet();
				pictureBox1.Refresh();
			}
		}
		
		private void InitializeSheet()
		{
			_maxVertNumbersCount = _sheet.MaxVertNumbersCount;
			_maxHorizNumbersCount = _sheet.MaxHorizNumbersCount;
			int width = (_maxHorizNumbersCount + _sheet.BrockHorizCount) * (1+BROCK_PIXEL) + 1;
			int height = (_maxVertNumbersCount + _sheet.BrockVertCount) * (1+BROCK_PIXEL) + 1;
			_bitmap = new Bitmap(width, height);
			Graphics g = Graphics.FromImage(_bitmap);
			g.FillRectangle(Brushes.White, 0, 0, width, height);
			pictureBox1.Image = _bitmap;
			pictureBox1.Width = width;
			pictureBox1.Height = height;
			if(width > Screen.GetWorkingArea(this).Width ||
			   height > Screen.GetWorkingArea(this).Height){
				panel1.AutoSize = false;
				panel1.Width = Screen.GetWorkingArea(this).Width-20;
				panel1.Height = Screen.GetWorkingArea(this).Height-100;
			}
			else{
				panel1.AutoSize = false;
				this.ClientSize = new Size(width, height);
				panel1.Width = width;
				panel1.Height = height;
			}
		}
		
		private void DisplaySheet()
		{
			Graphics g = Graphics.FromImage(_bitmap);
			DisplayVertNumbers(g);
			DisplayHorizNumbers(g);
			DisplayLines(g);
		}
		
		private void DisplayNumber(Graphics g, Rectangle rect, ContinuityNumber number)
		{
			g.DrawString(number.Number.ToString(), _font,
			             number.Fixed ? Brushes.LightGray : Brushes.Black,
			             rect, centeringFormat);
			pictureBox1.Invalidate(new Region(rect));
		}
		
		private void DisplayVertNumber(Graphics g, int x, int index)
		{
			ContinuityNumber number = _sheet.VertNumbers[x][index];
			Rectangle rect = new Rectangle(
				(_maxHorizNumbersCount+x)*(1+BROCK_PIXEL),
				(_maxVertNumbersCount - _sheet.VertNumbers[x].Length + index)*(1+BROCK_PIXEL),
				BROCK_PIXEL, BROCK_PIXEL);
			DisplayNumber(g, rect, number);
		}
		
		private void DisplayHorizNumber(Graphics g, int y, int index)
		{
			ContinuityNumber number = _sheet.HorizNumbers[y][index];
			Rectangle rect = new Rectangle(
				(_maxHorizNumbersCount - _sheet.HorizNumbers[y].Length + index)*(1+BROCK_PIXEL),
				(_maxVertNumbersCount+y)*(1+BROCK_PIXEL),
				BROCK_PIXEL, BROCK_PIXEL);
			DisplayNumber(g, rect, number);
		}
		
		private void DisplayVertNumbers(Graphics g)
		{
			int x=0;
			foreach(ContinuityNumber[] numbers in _sheet.VertNumbers){
				for(int index=0; index<numbers.Length; index++){
					DisplayVertNumber(g, x, index);
				}
				x++;
			}
		}
		
		private void DisplayHorizNumbers(Graphics g)
		{
			int y=0;
			foreach(ContinuityNumber[] numbers in _sheet.HorizNumbers){
				for(int index=0; index<numbers.Length; index++){
					DisplayHorizNumber(g, y, index);
				}
				y++;
			}
		}
		
		private void DisplayLines(Graphics g)
		{
			int x1, x2;
			int y1, y2;
			Pen penGray = new Pen(Color.LightGray, 1F);
			Pen penBlack = new Pen(Color.Gray, 1F);
			x1 = _maxHorizNumbersCount*(1+BROCK_PIXEL);
			x2 = (_maxHorizNumbersCount+_sheet.BrockHorizCount)*(1+BROCK_PIXEL);
			for(int stepY=0; stepY<_maxVertNumbersCount; stepY++){
				y1 = stepY*(1+BROCK_PIXEL);
				g.DrawLine(stepY==0 ? penBlack : penGray, x1, y1, x2, y1);
			}
			for(int stepY=0; stepY<=_sheet.BrockVertCount; stepY++){
				y1 = (_maxVertNumbersCount+stepY)*(1+BROCK_PIXEL);
				g.DrawLine((stepY==0 || stepY==_sheet.BrockVertCount || (stepY%5)==0) ? penBlack : penGray,
				           0, y1, x2, y1);
			}
			y1 = _maxVertNumbersCount*(1+BROCK_PIXEL);
			y2 = (_maxVertNumbersCount+_sheet.BrockVertCount)*(1+BROCK_PIXEL);
			for(int stepX=0; stepX<_maxHorizNumbersCount; stepX++){
				x1 = stepX*(1+BROCK_PIXEL);
				g.DrawLine(stepX==0 ? penBlack : penGray, x1, y1, x1, y2);
			}
			for(int stepX=0; stepX<=_sheet.BrockHorizCount; stepX++){
				x1 = (_maxHorizNumbersCount+stepX)*(1+BROCK_PIXEL);
				g.DrawLine((stepX==0 || stepX==_sheet.BrockHorizCount || (stepX%5)==0) ? penBlack : penGray,
				           x1, 0, x1, y2);
			}
		}
		
		public void UpdateBrockState(int x, int y)
		{
			UpdateBrockStateDelegate method = new UpdateBrockStateDelegate(DisplayBrockState);
			if(this.Visible){
				Invoke(method, new object[]{ x, y });
			}
		}
		
		private void DisplayBrockState(int x, int y)
		{
			if(x < _sheet.BrockHorizCount &&
			   y < _sheet.BrockVertCount){
				Brock brock = _sheet.Brocks[x, y];
				Graphics g = Graphics.FromImage(_bitmap);
				Rectangle rect = new Rectangle(
					(_maxHorizNumbersCount+x)*(1+BROCK_PIXEL)+1,
					(_maxVertNumbersCount+y)*(1+BROCK_PIXEL)+1,
					BROCK_PIXEL, BROCK_PIXEL);
				switch(brock.State){
					case BrockState.Fill:
						g.FillRectangle(Brushes.Black, rect);
						break;
					case BrockState.None:
						g.FillRectangle(Brushes.White, rect);
						Pen pen = new Pen(Color.Black, 1F);
						g.DrawLine(pen, rect.Left, rect.Top, rect.Right-1, rect.Bottom-1);
						g.DrawLine(pen, rect.Right-1, rect.Top, rect.Left, rect.Bottom-1);
						break;
					default:
						g.FillRectangle(Brushes.White, rect);
						break;
				}
				pictureBox1.Invalidate(new Region(rect));
				if(brock.VertNumIndex >= 0){
					DisplayVertNumber(g, x, brock.VertNumIndex);
				}
				if(brock.HorizNumIndex >= 0){
					DisplayHorizNumber(g, y, brock.HorizNumIndex);
				}
			}
		}
		
		void ToolStripButtonStepClick(object sender, EventArgs e)
		{
			DoStep(true);
		}
		
		private void DoStep(bool stepExecute)
		{
			if(_sheet != null && !IsThreadRunning()){
				int step;
				if(!int.TryParse(toolStripTextBoxStep.Text, out step)){
					step = 1;
					toolStripTextBoxStep.Text = "1";
				}
				Explainer explainer = new Explainer() {
					Sheet = _sheet,
					Step = stepExecute ? step : -1,
					Form = this,
				};
				thread = new Thread(new ThreadStart(explainer.Start));
				thread.Start();
				EnableToolbarButtons();
			}
		}
		
		private void EnableToolbarButtons()
		{
			bool enabled = true;
			if(thread != null && thread.IsAlive){
				enabled = false;
			}
			toolStripButtonOpen.Enabled = enabled;
			toolStripButtonStep.Enabled = enabled;
			toolStripButtonAll.Enabled = enabled;
			toolStripButtonSave.Enabled = enabled;
			toolStripButtonPrint.Enabled = enabled;
		}
		
		private bool IsThreadRunning()
		{
			return thread != null && thread.ThreadState == ThreadState.Running;
		}
		
		void MainFormFormClosed(object sender, FormClosedEventArgs e)
		{
			if(IsThreadRunning()){
				thread.Join();
			}
		}
		
		void ToolStripTextBoxStepKeyPress(object sender, KeyPressEventArgs e)
		{
			if(e.KeyChar == 0x0d){
				DoStep(true);
				e.Handled = true;
			}
		}
		
		void ToolStripButtonAllClick(object sender, EventArgs e)
		{
			DoStep(false);
		}

		void ToolStripButtonPrintClick(object sender, EventArgs e)
		{
			if(_sheet != null && !IsThreadRunning()){
				PrintDialog dialog = new PrintDialog();
				dialog.Document = printDocument1;
				if(dialog.ShowDialog() == DialogResult.OK){
					printDocument1.Print();
				}
			}
		}
		
		void PrintDocument1PrintPage(object sender, System.Drawing.Printing.PrintPageEventArgs e)
		{
			Rectangle rect = new Rectangle(e.MarginBounds.X, e.MarginBounds.Y, e.MarginBounds.Width, e.MarginBounds.Height);
			float ratio = (float)_bitmap.Width/_bitmap.Height;
			if(rect.Height*ratio > rect.Width){
				rect.Height = (int)(rect.Width/ratio);
			}
			else{
				rect.Width = (int)(rect.Height*ratio);
			}
			e.Graphics.DrawImage(_bitmap, rect);
			e.HasMorePages = false;
		}
		
		void PictureBox1MouseMove(object sender, MouseEventArgs e)
		{
			if(_sheet != null){
				int x = e.X/(1+BROCK_PIXEL) - _maxHorizNumbersCount;
				int y = e.Y/(1+BROCK_PIXEL) - _maxVertNumbersCount;
				if(x >= 0 && x < _sheet.BrockHorizCount &&
				   y >= 0 && y < _sheet.BrockVertCount &&
				   (x != _mouseX || y != _mouseY)){
					_mouseX = x;
					_mouseY = y;
					Brock brock = _sheet.Brocks[x, y];
					toolStripLabelDebug.Text = string.Format(
						"({0},{1})H{2}V{3}",
						x, y,
						brock.HorizNumIndex < 0 ? "?":brock.HorizNumIndex.ToString(),
						brock.VertNumIndex < 0 ? "?":brock.VertNumIndex.ToString());
				}
			}
		}
		
		
		void ToolStripButtonSaveClick(object sender, EventArgs e)
		{
			SaveFileDialog dialog = new SaveFileDialog();
			dialog.RestoreDirectory = true;
			dialog.Filter = "画像ファイル(*.png)|*.png";
			if(dialog.ShowDialog() == DialogResult.OK){
				pictureBox1.Image.Save(dialog.FileName, ImageFormat.Png);
			}
		}
		
		void Timer1Tick(object sender, EventArgs e)
		{
			EnableToolbarButtons();
		}
	}
}
