﻿/*
 * Created by SharpDevelop.
 * User: banana
 * Date: 2013/10/25
 * Time: 10:46
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Collections.Generic;
using System.Text;
using log4net;

namespace PicLogic.Processors
{
	/// <summary>
	/// Description of BaseProcessor.
	/// </summary>
	public abstract class BaseProcessor
	{
		protected ILog log;
		
		protected bool IsVertical { get; set; }
		
		public BaseProcessor()
		{
			log = LogManager.GetLogger(this.GetType());
		}
		
		public int[] Do(ContinuityNumber[] nums, Brock[] line, bool vertical)
		{
			IsVertical = vertical;
			return Execute(nums, line);
		}
		
		protected abstract int[] Execute(ContinuityNumber[] nums, Brock[] line);
		
		protected int CountSameState(Brock[] line, int startIndex)
		{
			int result = 0;
			if(startIndex < line.Length){
				BrockState state = line[startIndex].State;
				for(int index=startIndex; index<line.Length; index++){
					if(line[index].State != state){
						break;
					}
					result++;
				}
			}
			return result;
		}
		
		protected int FindStateEqual(Brock[] line, int startIndex, BrockState state)
		{
			for(int index=startIndex; index<line.Length; index++){
				if(line[index].State == state){
					return index;
				}
			}
			return line.Length;
		}
		
		protected int FindStateEqual(Brock[] line, int startIndex, int maxLength, BrockState state)
		{
			int count=0;
			for(int index=startIndex; index<line.Length && count<maxLength; index++, count++){
				if(line[index].State == state){
					return index;
				}
			}
			return line.Length;
		}
		
		protected int FindStateNotEqual(Brock[] line, int startIndex, BrockState state)
		{
			for(int index=startIndex; index<line.Length; index++){
				if(line[index].State != state){
					return index;
				}
			}
			return line.Length;
		}
		
		protected int FindStateNotEqual(Brock[] line, int startIndex, int maxLength, BrockState state)
		{
			int count=0;
			for(int index=startIndex; index<line.Length && count<maxLength; index++, count++){
				if(line[index].State == state){
					return index;
				}
			}
			return line.Length;
		}
		
		protected bool SetState(Brock[] line, int index, BrockState state)
		{
			if(index >= 0 && index < line.Length){
				if(line[index].State != state){
					line[index].State = state;
					log.DebugFormat("{0}:{1}", state.ToString(), index);
					return true;
				}
			}
			return false;
		}
		
		protected int FindNumIndexEqual(Brock[] line, int startIndex, int numIndex)
		{
			for(int index=startIndex; index<line.Length; index++){
				if(GetNumIndex(line[index]) == numIndex){
					return index;
				}
			}
			return line.Length;
		}
		
		protected int FindNumIndexEqualRev(Brock[] line, int startIndex, int numIndex)
		{
			for(int index=startIndex; index>=0; index--){
				if(GetNumIndex(line[index]) == numIndex){
					return index;
				}
			}
			return -1;
		}
		
		protected bool SetNumIndex(Brock[] line, int index, int numIndex)
		{
			if(GetNumIndex(line[index]) != numIndex){
				log.DebugFormat("numIndex={0}:{1}", numIndex, index);
				if(IsVertical){
					line[index].VertNumIndex = numIndex;
				}
				else{
					line[index].HorizNumIndex = numIndex;
				}
				return true;
			}
			return false;
		}
		
		protected bool SetStateAndNumIndex(Brock[] line, int index, BrockState state, int numIndex)
		{
			bool result = SetState(line, index, state);
			if(SetNumIndex(line, index, numIndex)){
				result = true;
			}
			return result;
		}
		
		protected int[] SetNumIndexToFill(Brock[] line, int startIndex, int numIndexValue)
		{
			List<int> result = new List<int>();
			for(int index=startIndex; index<line.Length && line[index].State== BrockState.Fill; index++){
				if(SetNumIndex(line, index, numIndexValue)){
					result.Add(index);
				}
			}
			return result.ToArray();
		}
		
		protected int GetNumIndex(Brock brock)
		{
			return IsVertical ? brock.VertNumIndex : brock.HorizNumIndex;
		}
		
		protected int Find(int[] array, int startIndex, int value)
		{
			for(int index=0; index<array.Length; index++){
				if(array[index] == value){
					return index;
				}
			}
			return array.Length;
		}
		
		protected int FindNumber(ContinuityNumber[] nums, int index)
		{
			for(int i=0; i<nums.Length; i++){
				if(nums[i].Index == index){
					return i;
				}
			}
			return nums.Length;
		}

		protected int[] NewIndexArray(int length, int initialValue)
		{
			int[] result = new int[length];
			for(int index=0; index<length; index++){
				result[index] = initialValue;
			}
			return result;
		}
		
		/// <summary>
		/// マッチする可能性のある数インデックスのリストを取得します。
		/// </summary>
		/// <param name="nums"></param>
		/// <param name="line"></param>
		/// <param name="position"></param>
		/// <returns></returns>
		protected int[] GetMatchNumIndexes(ContinuityNumber[] nums, Brock[] line, int position)
		{
			List<int> result = new List<int>();
			if(position < line.Length && line[position].State == BrockState.Fill){
				int currentFillCount = CountSameState(line, position);
				int[] leftSet = TrySetLeft(nums, line);
				int[] rightSet = TrySetRight(nums, line);
				DebugLog("前詰め", leftSet);
				DebugLog("後詰め", rightSet);
				int minIndex = FindNumber(nums, rightSet[position]);
				int maxIndex = FindNumber(nums, leftSet[position]);
				if(minIndex >= nums.Length || maxIndex >= nums.Length){
					log.Error("何かが間違っている");
				}
				else{
					for(int numIndex=minIndex; numIndex<=maxIndex; numIndex++){
						if(nums[numIndex].Number >= currentFillCount){
							result.Add(nums[numIndex].Index);
						}
					}
				}
			}
			DebugLog("数indexの候補", result.ToArray());
			return result.ToArray();
		}
		
		/// <summary>
		/// 左から詰めて並べてみます。
		/// </summary>
		/// <param name="nums"></param>
		/// <param name="line"></param>
		/// <returns></returns>
		protected int[] TrySetLeft(ContinuityNumber[] nums, Brock[] line)
		{
			// TODO 詰めすぎて、塗り潰しが余ったらどうするか
			int[] result = NewIndexArray(line.Length, -1);
			int index = 0;
			for(int numIndex=0; numIndex<nums.Length; numIndex++){
				ContinuityNumber num = nums[numIndex];
				if(num.FillCount > 0){
					int startIndex = index;
					while(index < line.Length && GetNumIndex(line[index]) != num.Index){ index++; }
					// 連続した塗り潰しの個数を数える
					int fillCount = CountSameState(line, index);
//					log.DebugFormat("連続個数={0}, 開始位置={1}, 対象の数={2}", fillCount, index, num.Number);
					if(fillCount < num.Number){
						// 足りてない場合は、足りない数をnとして、最大n個まで空きかどうか戻ってみる
						index--;
						for(int back=0; back<num.Number-fillCount && index>=startIndex; back++, index--){
							if(line[index].State == BrockState.None){
								break;
							}
						}
						index++; /* <= これが仮置きの開始位置 */
						// 但し、直前が塗り潰しなら1マス空けなければならない
						if(index >= 1 && line[index-1].State == BrockState.Fill){
//							log.DebugFormat("直前が塗り潰し index={0}", index);
							index++;
						}
						// 更に、次の塗り潰しとくっついてしまう場合は、それを含めるようにする
						if(index+num.Number < line.Length && line[index+num.Number].State == BrockState.Fill){
//							log.DebugFormat("直後が塗り潰し index={0}, number={1}", index, num.Number);
							index += CountSameState(line, index+num.Number);
//							log.DebugFormat("ずらした結果 index={0}", index);
						}
					}
					else if(fillCount > num.Number){
						log.ErrorFormat("既に{0}個以上並んでいます。", num.Number);
					}
					else{
						// 確定してる
					}
					// 仮置きする
//					log.DebugFormat("仮置き開始位置={0}", index);
					for(int count=0; count<num.Number && index<line.Length; count++, index++){
						if(IsSameNumberOrUnknown(line[index], num.Index)){
							result[index] = num.Index;
						}
					}
					index++;
				}
				else{
//					log.DebugFormat("曖昧な配置位置検索 開始位置={0}, 対象の数={1}", index, num.Number);
					while(index<line.Length){
						index = FindStateNotEqual(line, index, BrockState.None);
						int unknownCount = FindStateEqual(line, index, BrockState.None) - index;
						//bool fillAvail = FindStateEqual(line, index, BrockState.Fill) - index < unknownCount;
						if(unknownCount >= num.Number){
							// 連続した領域を見つけた
							// でも、次の塗り潰しとくっついてしまう場合は、それを含めるようにする
							if(index+num.Number < line.Length && line[index+num.Number].State == BrockState.Fill){
//								log.DebugFormat("直後が塗り潰し index={0}, number={1}", index, num.Number);
								index += CountSameState(line, index+num.Number);
//								log.DebugFormat("ずらした結果 index={0}", index);
							}
							for(int count=0; count<num.Number; count++){
								result[index++] = num.Index;
							}
							index++;
							break;
						}
						else{
							index += unknownCount;
						}
					}
				}
			}
			return result;
		}

		private bool IsSameNumberOrUnknown(Brock brock, int numIndex)
		{
			return brock.State == BrockState.Unknown ||
				(brock.State == BrockState.Fill && GetNumIndex(brock) == numIndex);
		}
		
		/// <summary>
		/// 右から詰めて並べてみます。
		/// </summary>
		/// <param name="nums"></param>
		/// <param name="line"></param>
		/// <returns></returns>
		protected int[] TrySetRight(ContinuityNumber[] nums, Brock[] line)
		{
			Array.Reverse(nums);
			Array.Reverse(line);
			int[] result = TrySetLeft(nums, line);
			Array.Reverse(nums);
			Array.Reverse(line);
			Array.Reverse(result);
			return result;
		}
		
		protected int[] ReverseBrockIndexValue(IEnumerable<int> indexes, int length)
		{
			List<int> result = new List<int>();
			foreach(int index in indexes){
				result.Add(length-index-1);
			}
			return result.ToArray();
		}
		
		protected void DebugLog(string message, int[] intArray)
		{
			if(log.IsDebugEnabled){
				StringBuilder sb = new StringBuilder();
				foreach(int i in intArray){
					sb.AppendFormat("{0},", i.ToString());
				}
				log.DebugFormat("{0}:{1}", message, sb.ToString());
			}
		}

		protected void DebugLog(string message, Brock[] line)
		{
			if(log.IsDebugEnabled){
				StringBuilder sb = new StringBuilder();
				foreach(Brock brock in line){
					sb.AppendFormat("{0}{1},", brock.State.ToString()[0], GetNumIndex(brock));
				}
				log.DebugFormat("{0}:{1}", message, sb.ToString());
			}
		}
	}
}
