﻿/*
 * Created by SharpDevelop.
 * User: bana
 * Date: 2013/11/01
 * Time: 11:53
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Collections.Generic;

namespace PicLogic.Processors
{
	/// <summary>
	/// 隙間なく並んだ塗りつぶしとXがあれば、塗りつぶしの個数を複数で推測する
	/// 1 2 5 2 1 9 1|____##X###_____ なら、2個以上と3個以上が並んでいるので、
	/// 1 2 5 2 1 9 1|___X##X#####X__ と確定できる
	/// </summary>
	public class CheckNumberMultiProcessor : BaseProcessor
	{
		public CheckNumberMultiProcessor()
		{
		}
		
		protected override int[] Execute(ContinuityNumber[] nums, Brock[] line)
		{
			List<int> result = new List<int>();
			List<FillData> fills = new List<CheckNumberMultiProcessor.FillData>();
			for(int index=0; index<line.Length;){
				// 数が確定していない塗りつぶしを見つける
				index = FindStateEqual(line, index, BrockState.Fill);
				if(index >= line.Length){
					continue;
				}
				int count = CountSameState(line, index);
				// 右端が X で止まっているか
				if(IsStop(line, index+count)){
					fills.Clear();
					fills.Add(new FillData {
					          	StartIndex=index,
					          	Count=count,
					          	Quoted=IsStop(line, index-1),
					          	NumIndex=GetNumIndex(line[index])
					          });
					bool quoted=true;
					while(quoted){
						index = FindStateNotEqual(line, index+count, BrockState.None);
						// X の次がすぐ塗りつぶしかどうか
						if(index < line.Length && line[index].State == BrockState.Fill){
							count = CountSameState(line, index);
							quoted = IsStop(line, index+count);
							fills.Add(new FillData {
							          	StartIndex=index,
							          	Count=count,
							          	Quoted=quoted,
							          	NumIndex=GetNumIndex(line[index])});
						}
						else{
							quoted = false;
						}
					}
					int startNumIndex = FindNumbers(nums, fills);
					if(startNumIndex < nums.Length){
						result.AddRange(SetNumbers(nums, line, fills, startNumIndex));
					}
				}
				else{
					index += count;
				}
			}
			return result.ToArray();
		}
		
		private int FindNumbers(ContinuityNumber[] nums, List<FillData> fills)
		{
			if(fills.Count >= 2){
				List<int> result = new List<int>();
				for(int numIndex=0; numIndex<nums.Length; numIndex++){
					if(IsMatch(nums[numIndex], fills[0])){
						bool unmatch = false;
						int fillIndex;
						for(fillIndex=1; fillIndex<fills.Count && numIndex+fillIndex<nums.Length && !unmatch; fillIndex++){
							if(!IsMatch(nums[numIndex+fillIndex], fills[fillIndex])){
								unmatch = true;
							}
						}
						if(fillIndex >= fills.Count && !unmatch){
							result.Add(numIndex);
						}
					}
				}
				log.DebugFormat("matchCount={0}", result.Count);
				if(result.Count == 1){
					// マッチするのがひとつだけの場合はそれで確定できる
					return result[0];
				}
			}
			return nums.Length;
		}
		
		private bool IsMatch(ContinuityNumber num, FillData fill)
		{
			if(!num.Fixed){
				log.DebugFormat("match? {0}:{1}", num.ToString(), fill.ToString());
				if(fill.NumIndex < 0 || fill.NumIndex==num.Index){
					if(fill.Quoted){
						return num.Number == fill.Count;
					}
					return num.Number >= fill.Count;
				}
			}
			return false;
		}
		
		private int[] SetNumbers(ContinuityNumber[] nums, Brock[] line,
		                         List<FillData> fills, int numIndex)
		{
			List<int> result = new List<int>();
			foreach(FillData fill in fills){
				result.AddRange(SetNumIndexToFill(line, fill.StartIndex, nums[numIndex].Index));
				numIndex++;
			}
			return result.ToArray();
		}
		
		private bool IsStop(Brock[] line, int index)
		{
			if(index < 0 || index >= line.Length){
				return true;
			}
			return line[index].State == BrockState.None;
		}
		
		private class FillData {
			public int StartIndex { get; set; }
			public int Count { get; set; }
			public bool Quoted { get; set; }
			public int NumIndex { get; set; }
			
			public override string ToString()
			{
				return string.Format("[FillData StartIndex={0}, Count={1}, Quoted={2}, NumIndex={3}]", StartIndex, Count, Quoted, NumIndex);
			}

		}
	}
}
