﻿/*
 * Created by SharpDevelop.
 * User: banana
 * Date: 2013/10/25
 * Time: 10:58
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Collections.Generic;

namespace PicLogic.Processors
{
	/// <summary>
	/// Description of CheckIndexProcessor.
	/// </summary>
	public class CheckNumberProcessor : BaseProcessor
	{
		public CheckNumberProcessor()
		{
		}
		
		protected override int[] Execute(ContinuityNumber[] nums, Brock[] line)
		{
			List<int> result = new List<int>();
			result.AddRange(SetSameIndex(nums, line));
			return result.ToArray();
		}
		
		private int GetAvailNumIndex(Brock[] line, int startIndex)
		{
			for(int index=startIndex; index<line.Length && line[index].State==BrockState.Fill; index++){
				int numIndex = GetNumIndex(line[index]);
				if(numIndex >= 0){
					return numIndex;
				}
			}
			return -1;
		}
		
		private int[] SetSameIndex(ContinuityNumber[] nums, Brock[] line)
		{
			List<int> result = new List<int>();
			int lastNumIndex = -1;
			DebugLog("Current", line);
			// 連続した塗り潰しに同じ数インデックスを割り当てる
			for(int index=0; index<line.Length; index++){
				index = FindStateEqual(line, index, BrockState.Fill);
				int numIndex = GetAvailNumIndex(line, index);
				if(numIndex >= 0){
					log.DebugFormat("既に数インデックスが設定済み:index={0}, numIndex={1}", index, numIndex);
				}
				else{
					// 推測
					int currentFillCount = CountSameState(line, index);
					int[] numIndexes = GetNumbersMoreThan(nums, currentFillCount);
					if(numIndexes.Length == 1){
						log.DebugFormat("連続した塗り潰し数{0}以上の数は唯一{1}しかないので確定", currentFillCount, nums[numIndexes[0]].Number);
						numIndex = numIndexes[0];
					}
					else{
						numIndexes = GetMatchNumIndexes(nums, line, index);
						if(numIndexes.Length == 1){
							// 確定
							numIndex = numIndexes[0];
						}
					}
				}
				if(numIndex >= 0){
					if(numIndex == lastNumIndex && index < line.Length){
						// 前回と同じ数インデックスならつなぐ
						for(int linkIndex=index-1; linkIndex>=0 && line[linkIndex].State == BrockState.Unknown; linkIndex--){
							log.DebugFormat("つなぐ index={0}", linkIndex);
							if(SetStateAndNumIndex(line, linkIndex, BrockState.Fill, numIndex)){
								result.Add(linkIndex);
							}
						}
					}
					for(; index<line.Length && line[index].State==BrockState.Fill; index++){
						if(GetNumIndex(line[index]) != numIndex){
							log.DebugFormat("index={0}, numIndex={1}", index, numIndex);
							if(SetNumIndex(line, index, numIndex)){
								result.Add(index);
							}
						}
					}
				}
				lastNumIndex = numIndex;
				index = FindStateNotEqual(line, index, BrockState.Fill);
			}
			return result.ToArray();
		}
		
		private int[] GetNumbersMoreThan(ContinuityNumber[] nums, int count)
		{
			List<int> result = new List<int>();
			foreach(ContinuityNumber num in nums){
				if(!num.Fixed && num.Number >= count){
					result.Add(num.Index);
				}
			}
			return result.ToArray();
		}
	}
}
