﻿/*
 * Created by SharpDevelop.
 * User: banana
 * Date: 2013/10/25
 * Time: 10:57
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Collections.Generic;

namespace PicLogic.Processors
{
	/// <summary>
	/// 4 1|_#___ → |_###_
	/// 4 1|_X_#__ → |_X_###
	/// </summary>
	public class FromEdgeProcessor : BaseProcessor
	{
		public FromEdgeProcessor()
		{
		}
		
		protected override int[] Execute(ContinuityNumber[] nums, Brock[] line)
		{
			List<int> result = new List<int>();
			// 先頭から
			result.AddRange(FromEdge(nums, line));
			// 末尾から
			Array.Reverse(nums);
			Array.Reverse(line);
			result.AddRange(ReverseBrockIndexValue(FromEdge(nums, line), line.Length));
			Array.Reverse(nums);
			Array.Reverse(line);
			return result.ToArray();
		}
		
		private int SkipFixedFill(ContinuityNumber[] nums, Brock[] line, int numIndex, int index, List<int> updateIndexes)
		{
			index = FindStateEqual(line, index, BrockState.Fill);
			index += nums[numIndex].Number;
			// 終了に X が付いていなければ付ける
			if(index < line.Length){
				if(line[index].State == BrockState.Unknown){
					if(SetState(line, index, BrockState.None)){
						updateIndexes.Add(index);
					}
				}
			}
			return index;
		}
		private int[] FromEdge(ContinuityNumber[] nums, Brock[] line)
		{
			List<int> result = new List<int>();
			DebugLog("状態", line);
			bool fix = true;
			int numIndex = 0;
			for(int startIndex=0; startIndex<line.Length && numIndex<nums.Length; ){
				log.DebugFormat("startIndex={0}, numIndex={1}, fix={2}", startIndex, numIndex, fix);
				if(nums[numIndex].Fixed){
					// 既に確定
					log.DebugFormat("既に確定 numIndex={0}", numIndex);
					if(FindNumIndexEqual(line, startIndex, nums[numIndex].Index) >= line.Length){
						// 通り過ぎている
						log.DebugFormat("でも通り過ぎている");
						numIndex++;
						continue;
					}
					startIndex = SkipFixedFill(nums, line, numIndex, startIndex, result);
					numIndex++;
					fix = true;
				}
				else{
					// X ではない所を開始位置とする
					startIndex = FindStateNotEqual(line, startIndex, BrockState.None);
					// 空きの長さ
					int unknownCount = CountSameState(line, startIndex);
					bool fuzzy = false;
					// 最初の塗り潰しを見つける
					int fillStartIndex = FindStateEqual(line, startIndex, nums[numIndex].Number, BrockState.Fill);
					// 最初の塗り潰しからの追加塗り潰し個数を計算
					int fillLength = nums[numIndex].Number - (fillStartIndex-startIndex);
					if(fix){
						bool nextFill = (startIndex+unknownCount < line.Length &&
						                 line[startIndex+unknownCount].State == BrockState.Fill);
						if(nums[numIndex].Number == 1 && unknownCount == 2 && nextFill){
							// 1 2 3|__# なら、#の前にXが入り、|_X# になる
							log.DebugFormat("数字1に対して2個空白があるのでXを付ける. index={0}", startIndex);
							int index=startIndex+1;
							if(SetState(line, index, BrockState.None)){
								result.Add(index);
							}
						}
						else if(nums[numIndex].Number == unknownCount && nextFill){
							// 2 3|__# (数字と空白数が同じ)なら、先頭にXが入り、|X_# になる
							log.DebugFormat("数字{0}に対して同数の空白があるので先頭にXを付ける. index={1}", unknownCount, startIndex);
							if(SetState(line, startIndex, BrockState.None)){
								result.Add(startIndex);
							}
						}
					}
					if(fillStartIndex < line.Length){
						if(!fix){
							int brockNumIndex = GetNumIndex(line[fillStartIndex]);
							// 塗り潰し連続個数を確認
							int currentFillCount = CountSameState(line, fillStartIndex);
							log.DebugFormat("currentFillCount={0}", currentFillCount);
							if(brockNumIndex < 0){
								// 対象の数字が解らないなら条件に一致する候補を取得
								int[] numIndexes = GetMatchNumIndexes(nums, line, fillStartIndex);
								if(numIndexes.Length == 1){
									log.DebugFormat("match only 1. numIndex={0}", numIndexes[0]);
									numIndex = FindNumber(nums, numIndexes[0]);
									fillLength = nums[numIndex].Number - (fillStartIndex-startIndex);
								}
								else{
									fillLength = GetMinNumber(nums, numIndexes) - (fillStartIndex-startIndex);
									fuzzy = true;
								}
							}
							else if(brockNumIndex != nums[numIndex].Index){
								numIndex = FindNumber(nums, brockNumIndex);
								fillLength = nums[numIndex].Number - (fillStartIndex-startIndex);
							}
						}
					}
					fix = false;
					log.DebugFormat("fillStartIndex={0}, fillLength={1}", fillStartIndex, fillLength);
					if(fillLength > 0 && fillStartIndex+fillLength < line.Length){
						// 塗り潰す
						startIndex = SetFill(nums, line, numIndex, fillStartIndex, fillLength, fuzzy, result);
						numIndex++;
					}
				}
				// TODO この辺りの処理は怪しい(numIndexとstartIndexの対応が取れない)
				// 次の X を探す
				startIndex = FindStateEqual(line, startIndex, BrockState.None);
				// 次の X 以外を探す
				startIndex = FindStateNotEqual(line, startIndex, BrockState.None);
			}
			return result.ToArray();
		}
		
		private int SetFill(ContinuityNumber[] nums, Brock[] line, int numIndex, int startIndex, int length, bool fuzzy, List<int> updateIndexes)
		{
			int index = startIndex;
			for(index=startIndex; index<startIndex+length; index++){
				Brock brock = line[index];
				if(brock.State == BrockState.Unknown ||
				   (!fuzzy && GetNumIndex(brock) != nums[numIndex].Index)){
					SetState(line, index, BrockState.Fill);
					if(!fuzzy) SetNumIndex(line, index, nums[numIndex].Index);
					updateIndexes.Add(index);
				}
			}
			if(!fuzzy && nums[numIndex].Number == length && index < line.Length){
				Brock brock = line[index];
				if(brock.State == BrockState.Unknown){
					if(SetState(line, index, BrockState.None)){
						updateIndexes.Add(index);
					}
				}
			}
			return index;
		}
		
		private int GetMinNumber(ContinuityNumber[] nums, int[] numIndexes)
		{
			int result = -1;
			foreach(int numIndex in numIndexes){
				log.DebugFormat("{0}:{1}", numIndex, nums[numIndex].ToString());
				int number = nums[FindNumber(nums, numIndex)].Number;
				if(result<0 || result>number){
					result = number;
				}
			}
			if(result > 0){
				log.DebugFormat("min={0}", result);
			}
			return result;
		}
	}
}
